from pydantic.dataclasses import dataclass

from typing import List, Optional


@dataclass
class ColumnUDFConfig:
    """
    Attributes:
        function: BigQuery function to generate the additional column
        output_type: Bigquery datatype
    """

    function: str
    output_type: str



from airflow.exceptions import AirflowException

from dacite import Config
from dataclasses import dataclass, field
from pydantic import validator, root_validator

from datetime import datetime
from typing import List, Optional

from gcp_airflow_foundations.base_class.ods_metadata_config import (
    OdsTableMetadataConfig,
)
from gcp_airflow_foundations.enums.time_partitioning import TimePartitioning


@dataclass
class OdsTableConfig:
    """
    Attributes:
        ods_metadata : See OdsTableMetadataConfig class
        merge_type : Only SG_KEY_WITH_HASH is currently supported
        ods_table_time_partitioning: Partitioning for BigQuery table. One of DAY, HOUR, or MONTH. Must be provided if partition_column_name is not
        partition_column_name: Column to use for partitioning. Must be provided if ods_table_time_partitioning is not
    """

    ods_table_time_partitioning: Optional[TimePartitioning]  # DAY, HOUR, or MONTH
    partition_column_name: Optional[str]
    ods_metadata: OdsTableMetadataConfig = OdsTableMetadataConfig()
    merge_type: str = "SG_KEY_WITH_HASH"
    ods_table_time_partitioning: Optional[
        TimePartitioning
    ] = None  # DAY, HOUR, or MONTH
    partition_column_name: Optional[str] = None

    @root_validator(pre=True)
    def valid_config(cls, values):
        if values["ods_table_time_partitioning"] is not None:
            assert (
                values["partition_column_name"] is not None
            ), "Please select a column for time partitioning."
        return values



import pandas as pd

from gcp_airflow_foundations.base_class.dlp_source_config import PolicyTagConfig


def get_dlp_results_sql(project_id, dataset_id, table_id, min_count=0) -> str:
    """ Generate sql to query the DLP results table: https://cloud.google.com/dlp/docs/querying-findings
        and counts the number of finding in each field,info-type, likelihood grouping
    :param project_id: Project Id
    :param dataset_id: Dataset
    :param table_id: Table
    :return:a sql query that generates a result set with the following columns [field_name, info_type_name, likelihood, count_total]
    """

    return """SELECT
        locations.record_location.field_id.name AS field_name,
        info_type.name as info_type_name,
        likelihood as likelihood,
        COUNT(*) AS count_total
    FROM {}.{}.{},
        UNNEST(location.content_locations) AS locations
    GROUP BY
        locations.record_location.field_id.name,
        info_type.name,
        likelihood
    HAVING count_total > {}
    """.format(
        project_id, dataset_id, table_id, min_count
    )


def results_to_bq_policy_tags(
    project_id, fields, poicy_tag_config: PolicyTagConfig
) -> list:
    """ Generates a tag to be applied for each column (if any)
    :param fields: A fields as generated by a query from get_dlp_results_sql
    :param tag: The tag to apply
    :return: A list of columns with am updated policyTags as per  https://cloud.google.com/bigquery/docs/reference/rest/v2/tables#TableSchema
 """
    df = pd.DataFrame(
        fields, columns=["field_name", "info_type_name", "likelihood", "count_total"]
    )
    tags = []
    unique = df.drop_duplicates(subset=["field_name"])
    for field in unique["field_name"]:
        tags.append(
            {
                "name": field,
                "policy_tags": {
                    "names": [policy_tag_name(fields, project_id, poicy_tag_config)]
                },
            }
        )
    return tags


def policy_tag_name(field, project_id, poicy_tag_config: PolicyTagConfig):
    return f"projects/{project_id}/locations/{poicy_tag_config.location}/taxonomies/{poicy_tag_config.taxonomy}/policyTags/{poicy_tag_config.tag}"



from gcp_airflow_foundations.operators.api.hooks.dataform_hook import DataformHook
from airflow.sensors.base import BaseSensorOperator
# from airflow.utils.decorators import apply_defaults
from airflow.exceptions import AirflowException
from typing import Any
import logging


class DataformSensor(BaseSensorOperator):
    # @apply_defaults
    def __init__(self, task_ids: str, dataform_conn_id='dataform_conn_id', **kwargs: Any):
        super().__init__(**kwargs)
        self.task_ids = task_ids
        self.hook = DataformHook(dataform_conn_id=dataform_conn_id)

    def poke(self, context):
        run_url = context['ti'].xcom_pull(key='return_value', task_ids=[self.task_ids])
        response = self.hook.check_job_status(run_url[0])

        if response == 'SUCCESSFUL':
            logging.info(f"Dataform run completed: SUCCESSFUL see run logs at {run_url}")
            return True
        elif response == 'RUNNING':
            logging.info(f"Dataform job running. {run_url}")
            return False
        else:
            raise AirflowException(f"Dataform run {response}: see run logs at {run_url}")



from enum import Enum


class TemplateIngestionMode(Enum):
    INGEST_ALL = "INGEST_ALL"
    INGEST_BY_TABLE_NAMES = "INGEST_BY_TABLE_NAMES"
    INGEST_BY_REGEX = "INGEST_BY_REGEX"


class TemplateDagCreationMode(Enum):
    SOURCE = "SOURCE"
    TABLE = "TABLE"



from typing import Optional, Sequence

from airflow.models import BaseOperator

from google.cloud.bigquery.dataset import Dataset
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook


class CustomBigQueryCreateEmptyDatasetOperator(BaseOperator):
    """
    This operator is used to create new dataset for your Project.
    :param project_id: The name of the project.
    :param dataset_id: The id of dataset.
    :param location: The geographic location where the dataset should reside.
    :param exists_ok: If ``True``, ignore "already exists" errors.
    """

    template_fields: Sequence[str] = (
        'dataset_id',
        'project_id'
    )

    ui_color = "#5F86FF"

    def __init__(
        self,
        *,
        dataset_id: Optional[str] = None,
        project_id: Optional[str] = None,
        location: Optional[str] = None,
        exists_ok: bool = False,
        gcp_conn_id: str = 'google_cloud_default',
        **kwargs,
    ) -> None:

        self.dataset_id = dataset_id
        self.project_id = project_id
        self.location = location
        self.exists_ok = exists_ok
        self.gcp_conn_id = gcp_conn_id

        super().__init__(**kwargs)

    def execute(self, context) -> None:
        dataset_reference = {
            "datasetReference": {
                "datasetId": self.dataset_id,
                "projectId": self.project_id
            },
            "location": self.location}

        bq_hook = BigQueryHook(
            gcp_conn_id=self.gcp_conn_id,
            location=self.location
        )

        bq_client = bq_hook.get_client(
            project_id=self.project_id,
            location=self.location
        )

        dataset: Dataset = Dataset.from_api_repr(dataset_reference)

        self.log.info(
            'Creating dataset: %s in project: %s ',
            self.dataset_id,
            self.project_id
        )

        bq_client.create_dataset(dataset=dataset, exists_ok=self.exists_ok)

        self.log.info('Dataset created successfully.')



from dataclasses import fields
from os import X_OK
from urllib.parse import urlparse
import logging
from dacite import from_dict
from dataclasses import dataclass
from datetime import datetime
import shutil
import os
import glob

from airflow.operators.python_operator import PythonOperator
from airflow.operators.dummy import DummyOperator
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.sensors.python import PythonSensor
from airflow.models import Variable

from airflow.providers.sftp.hooks.sftp import SFTPHook
from gcp_airflow_foundations.operators.api.sensors.sftp_sensor import SFTPFilesExistenceSensor
from gcp_airflow_foundations.operators.api.sensors.sftp_prefix_sensor import SFTPFilesExistencePrefixSensor
from gcp_airflow_foundations.source_class.generic_file_source import GenericFileIngestionDagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder
from gcp_airflow_foundations.base_class.sftp_table_config import SFTPTableConfig
from gcp_airflow_foundations.common.sftp.helpers import save_id_to_file


class SFTPFileIngestionDagBuilder(GenericFileIngestionDagBuilder):
    """
    Builds DAGs to load files from an SFTP server to a BigQuery Table. Authenticates with the Airflow SFTP provider connection.

    For specifying files to be uploaded, there are 3 options:
        1) A metadata file is provided with templated daily file names (templating supports {{ ds }} and {{ TABLE_NAME }})
        2) The field templated_file_name is provided in the table config (the same templating applies - but only 1 file can be provided in this manner)
        3) A directory can be specified for .PARQUET files only in the table config

    If a metadata file is provided, it can be a fixed file, or can be a new file supplied daily.
    Airflow context variables are supported for the file naming, e.g.
        TABLE_METADATA_FILE_{{ ds }}.csv
    for a metadata file supplied daily.

    The format of the metadata file should be a csv with one column as follows:
        FILE_NAME_1
        ...
        FILE_NAME_N
    with all files to ingest

    If a directory for .PARQUET ingestion is provided, the field directory_prefix should point to the directory where
    the outward-most external partition sits, and the field flag_file_path should point to the path of the _SUCCESS file on the sftp server.
    A mod_time check occurs on the flag file to screen for daily ingestion on the sensor.
    """
    source_type = "SFTP"

    def flag_file_sensor(self, table_config, taskgroup):
        if "flag_file_path" in table_config.extra_options.get("file_table_config"):
            dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]
            flag_file_path = table_config.extra_options.get("file_table_config")["flag_file_path"]
            return self.get_sftp_sensor(table_config, taskgroup, dir_prefix, flag_file_path)
        else:
            return None

    def file_ingestion_task(self, table_config, taskgroup):
        """
        Ingest from SFTP
        """
        dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]
        if "flag_file_path" in table_config.extra_options.get("file_table_config"):
            flag_file_path = table_config.extra_options.get("file_table_config")["flag_file_path"]
        else:
            flag_file_path = ""
        gcs_bucket_prefix = self.config.source.extra_options["file_source_config"]["gcs_bucket_prefix"]
        bucket = self.config.source.extra_options["gcs_bucket"]
        return self.get_sftp_ingestion_operator(table_config, taskgroup, dir_prefix, gcs_bucket_prefix, flag_file_path, bucket)

    def get_file_list_task(self, table_config, taskgroup):
        return PythonOperator(
            task_id="get_file_list",
            op_kwargs={"table_config": table_config},
            python_callable=self.get_list_of_files,
            task_group=taskgroup)

    def get_list_of_files(self, table_config, **kwargs):
        # gcs_hook = GCSHook()
        airflow_date_template = self.config.source.extra_options["file_source_config"]["airflow_date_template"]
        if airflow_date_template == "ds":
            ds = kwargs["ds"]
        else:
            ds = kwargs["prev_ds"]
        ds = datetime.strptime(str(ds), "%Y-%m-%d").strftime(self.config.source.extra_options["file_source_config"]["date_format"])
        # XCom push the list of files
        # overwrite if in table_config
        dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]
        dir_prefix = dir_prefix.replace("{{ ds }}", ds)

        # gcs_bucket_prefix = self.config.source.extra_options["file_source_config"]["gcs_bucket_prefix"]

        if self.config.source.extra_options["file_source_config"]["source_format"] == "PARQUET":
            file_list = [dir_prefix]
            kwargs['ti'].xcom_push(key='file_list', value=file_list)
            return
        else:
            # bucket = self.config.source.extra_options["gcs_bucket"]
            if "metadata_file" in table_config.extra_options.get("file_table_config"):
                # metadata_file_name = table_config.extra_options.get("file_table_config")["metadata_file"]
                # metadata_file = gcs_hook.download(bucket_name=bucket, object_name=metadata_file_name, filename="metadata.csv")
                file_list = []
                with open('metadata.csv', newline='') as f:
                    for line in f:
                        file_list.append(line.strip())
            else:
                templated_file_name = self.config.source.extra_options["file_source_config"]["file_name_template"]
                templated_file_name = templated_file_name.replace("{{ TABLE_NAME }}", table_config.table_name)
                file_list = [templated_file_name]

            # support replacing files with current dates
            file_list[:] = [file.replace("{{ ds }}", ds) if "{{ ds }}" in file else file for file in file_list]

        kwargs['ti'].xcom_push(key='file_list', value=file_list)

    def file_sensor(self, table_config, taskgroup):
        """
        Returns an Airflow sensor that waits for the list of files specified by the metadata file provided.
        """
        # If we have a flag file for ingestion, there is no need to check the files specifically
        if "flag_file_path" in table_config.extra_options.get("file_table_config"):
            return None

        # bucket = self.config.source.extra_options["gcs_bucket"]
        table_name = table_config.table_name
        files_to_wait_for = "{{" + f"ti.xcom_pull(key='file_list', task_ids='{table_name}.ftp_taskgroup.get_file_list')" + "}}"
        timeout = self.config.source.extra_options["file_source_config"]["sensor_timeout"]
        sftp_conn = self.config.source.extra_options["sftp_source_config"]["sftp_connection_name"]
        file_prefix_filtering = timeout = self.config.source.extra_options["file_source_config"]["file_prefix_filtering"]
        dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]

        if file_prefix_filtering:
            if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
                key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
                return SFTPFilesExistencePrefixSensor(
                    task_id="wait_for_files_to_ingest",
                    path=dir_prefix,
                    prefixes=files_to_wait_for,
                    key_auth=True,
                    key_name=key_name,
                    task_group=taskgroup,
                    sftp_conn_id=sftp_conn,
                    timeout=timeout
                )
            else:
                return SFTPFilesExistencePrefixSensor(
                    task_id="wait_for_files_to_ingest",
                    path=dir_prefix,
                    prefixes=files_to_wait_for,
                    task_group=taskgroup,
                    sftp_conn_id=sftp_conn,
                    timeout=timeout
                )
        else:
            if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
                key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
                return SFTPFilesExistenceSensor(
                    task_id="wait_for_files_to_ingest",
                    dir_prefix=dir_prefix,
                    paths=files_to_wait_for,
                    key_auth=True,
                    key_name=key_name,
                    task_group=taskgroup,
                    sftp_conn_id=sftp_conn,
                    timeout=timeout
                )
            else:
                return SFTPFilesExistenceSensor(
                    task_id="wait_for_files_to_ingest",
                    dir_prefix=dir_prefix,
                    paths=files_to_wait_for,
                    task_group=taskgroup,
                    sftp_conn_id=sftp_conn,
                    timeout=timeout
                )

    def get_sftp_sensor(self, table_config, taskgroup, dir_prefix, flag_file_path):
        return PythonSensor(
            task_id="sftp_sensor",
            op_kwargs={"table_config": table_config,
                       "dir_prefix": dir_prefix,
                       "flag_file_path": flag_file_path},
            python_callable=self.sftp_flag_sensor,
            task_group=taskgroup
        )

    def sftp_flag_sensor(self, table_config, dir_prefix, flag_file_path, **kwargs):
        # Check if the flag file is dated with the execution date or later - if yes, then the files are ready for ingestion.
        # allows backfilling
        sftp_conn = self.config.source.extra_options["sftp_source_config"]["sftp_connection_name"]

        ds = kwargs["ds"]
        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            private_key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
            save_id_to_file(private_key_name)
        sftp_hook = SFTPHook(ssh_conn_id=sftp_conn)
        dir_prefix = dir_prefix.replace("{{ ds }}", ds)
        mod_time = sftp_hook.get_mod_time(flag_file_path)
        mod_time = datetime.strptime(mod_time, '%Y%m%d%H%M%S').strftime('%Y-%m-%d')
        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            if os.path.isfile("id_rsa"):
                os.remove("id_rsa")

        return ds <= mod_time

    def get_sftp_ingestion_operator(self, table_config, taskgroup, dir_prefix, gcs_bucket_prefix, flag_file_path, bucket):
        if not flag_file_path == "":
            return PythonOperator(
                task_id="load_sftp_to_gcs",
                op_kwargs={
                    "table_config": table_config,
                    "dir_prefix": dir_prefix,
                    "gcs_bucket_prefix": gcs_bucket_prefix,
                    "bucket": bucket,
                    "flag_file_path": flag_file_path},
                python_callable=self.load_sftp_to_gcs_flag,
                task_group=taskgroup
            )
        else:
            return PythonOperator(
                task_id="load_sftp_to_gcs",
                op_kwargs={
                    "table_config": table_config,
                    "dir_prefix": dir_prefix,
                    "gcs_bucket_prefix": gcs_bucket_prefix,
                    "bucket": bucket},
                python_callable=self.load_sftp_to_gcs,
                task_group=taskgroup
            )

    # flake8: noqa: C901
    def load_sftp_to_gcs(self, table_config, dir_prefix, gcs_bucket_prefix, bucket, **kwargs):
        ds = kwargs["ds"]
        ti = kwargs["ti"]
        sftp_conn = self.config.source.extra_options["sftp_source_config"]["sftp_connection_name"]

        if not os.path.isdir("sftp_data"):
            os.mkdir("sftp_data")
        my_path = "sftp_data"

        if gcs_bucket_prefix is None:
            gcs_bucket_prefix = ""
        if not gcs_bucket_prefix == "":
            gcs_bucket_prefix += "/"

        table_name = table_config.table_name

        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            private_key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
            save_id_to_file(private_key_name)
        sftp_hook = SFTPHook(sftp_conn)

        files_to_load = ti.xcom_pull(key='file_list', task_ids=f'{table_name}.ftp_taskgroup.get_file_list')

        if self.config.source.extra_options["file_source_config"]["file_prefix_filtering"]:
            file_name_prefixes = [f.split("/")[-1] for f in files_to_load]
            full_file_names = []
            dir = sftp_hook.list_directory(dir_prefix)
            for prefix in file_name_prefixes:
                for f in dir:
                    if f.startswith(prefix):
                        full_file_names.append(f)
                    break
            files_to_load = full_file_names

        for file in files_to_load:
            sftp_hook.retrieve_file(dir_prefix + "/" + file, my_path + "/" + file)

        destination_path_prefix = gcs_bucket_prefix + table_name + "/" + ds
        if "gcs_bucket_path_format_mode" in self.config.source.extra_options["file_source_config"]:
            date = datetime.strptime(str(ds), '%Y-%m-%d').strftime('%Y/%m/%d')
            destination_path_prefix = gcs_bucket_prefix + table_name + "/" + date

        gcs_hook = GCSHook()
        for file in files_to_load:
            obj = destination_path_prefix + "/" + file
            gcs_hook.upload(
                bucket_name=bucket,
                object_name=obj,
                filename=my_path + "/" + file)
        # delete data
        shutil.rmtree(my_path)

        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            if os.path.isfile("id_rsa"):
                os.remove("id_rsa")

    def load_sftp_to_gcs_flag(self, table_config, dir_prefix, gcs_bucket_prefix, bucket, flag_file_path, **kwargs):
        ds = kwargs["ds"]
        # ti = kwargs["ti"]

        airflow_date_template = self.config.source.extra_options["file_source_config"]["airflow_date_template"]

        if gcs_bucket_prefix is None:
            gcs_bucket_prefix = ""
        if not gcs_bucket_prefix == "":
            gcs_bucket_prefix += "/"

        # flag to check whether we are loading dates named with {{ prev_ds }} or {{ ds }}
        # landing zone table creation will still use {{ ds }}
        source_file_date = ds
        if airflow_date_template == "prev_ds":
            source_file_date = kwargs["prev_ds"]

        date_column = table_config.extra_options.get("sftp_table_config")["date_column"]
        full_dir_download = table_config.extra_options.get("sftp_table_config")["full_dir_download"]
        table_name = table_config.table_name
        destination_path_prefix = gcs_bucket_prefix + table_name + "/" + ds

        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            private_key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
            save_id_to_file(private_key_name)
        sftp_hook = SFTPHook()

        # Create directory and download data
        if not os.path.isdir("sftp_data"):
            os.mkdir("sftp_data")
        my_path = "sftp_data"
        os.chmod(my_path, 0o777)

        # get recursive sftp structure w/ full paths
        tree_map_prefix = dir_prefix
        upload_option = table_config.extra_options.get("file_table_config")["parquet_upload_option"]
        if not full_dir_download and upload_option == "BASH":
            tree_map_prefix += f"/{date_column}={source_file_date}"
        logging.info(tree_map_prefix)
        map = sftp_hook.get_tree_map(tree_map_prefix)
        if isinstance(map[0], list):
            map = map[0]

        # get list of files, create local directory structure, download
        files_to_upload = [x for x in map]
        # delete prefix up until the external partitions
        local_file_prefix_to_drop = ""
        if not files_to_upload == []:
            local_file_prefix_to_drop = my_path + "/" + dir_prefix + "/"

        gcs_hook = GCSHook()
        logging.info(files_to_upload)

        local_files_list = [my_path + "/" + file for file in files_to_upload]

        for remote_file in files_to_upload:
            file = my_path + "/" + remote_file
            path_to_create = file.rsplit('/', 1)[0]
            if not os.path.exists(path_to_create):
                folders = path_to_create.split("/")
                for i in range(len(folders)):
                    folder_to_create = "/".join(folders[0:i + 1])
                    if not os.path.exists(folder_to_create):
                        os.mkdir(folder_to_create)
                        os.chmod(folder_to_create, 0o777)

            if not remote_file == flag_file_path:
                sftp_hook.retrieve_file(remote_file, file)
                # Get correct GCS file names to upload
                obj = destination_path_prefix + "/" + file.replace(local_file_prefix_to_drop, "")
                logging.info(obj)
                gcs_hook.upload(
                    bucket_name=bucket,
                    object_name=obj,
                    filename=file)
                os.remove(file)

        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            if os.path.isfile("id_rsa"):
                os.remove("id_rsa")

        # xcom push the prefix of external partitions, e.g. "par1=val1/par2=val/par3=val3"
        # hacky way to do it - find a cleaner way to fix it later

        if "partition_prefix" not in self.config.source.extra_options["sftp_source_config"]:
            partition_prefix = local_files_list[0].rsplit('/', 1)[0].replace(my_path + "/" + dir_prefix, "").lstrip("/")
            if "/" in partition_prefix:
                partition_prefix = partition_prefix.rsplit('/', 1)[0]
        else:
            # TEMPORARY hack to fix ingestion. # TO REMOVE
            if table_config.extra_options.get("file_table_config")["parquet_upload_option"] == "BASH":
                partition_prefix = self.config.source.extra_options["sftp_source_config"]["partition_prefix"]
                partition_prefix.replace("date", table_config.extra_options.get("sftp_table_config")["date_column"])
                partition_prefix.replace("ds", ds)

        logging.info(partition_prefix)
        kwargs['ti'].xcom_push(key='partition_prefix', value=partition_prefix)

    def delete_gcs_files(self, table_config, taskgroup):
        pass

    def validate_extra_options(self):
        super().validate_extra_options()
        tables = self.config.tables
        for table_config in tables:
            # try and parse as SFTPTableConfig
            # sftp_table_config = from_dict(data_class=SFTPTableConfig, data=table_config.extra_options.get("sftp_table_config"))
            pass



from airflow.models.baseoperator import BaseOperator
from airflow.hooks.base import BaseHook
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook

# from airflow.utils.decorators import apply_defaults
from airflow.exceptions import AirflowException

from twilio.rest import Client
import logging
import os


def get_twilio_client(conn_id):
    try:
        connection = BaseHook.get_connection(conn_id)
        account_sid = connection.login
        auth_token = connection.password
        logging.info("Twilio connection established")
        return Client(account_sid, auth_token)
    except Exception as e:
        raise AirflowException(f"Twilio connection not established: {e}")


def get_bigquery_hook(bq_conn):
    try:
        hook = BigQueryHook(bq_conn, use_legacy_sql=False)
        # conn = hook.get_conn()
        logging.info("BigQuery connection established")
        return hook
    except Exception as e:
        raise AirflowException(f"Connection not estabilished: {str(e)}")


class BigQueryToCsv(BaseOperator):
    # @apply_defaults
    def __init__(self, *, bq_conn: str, query: str, file: str, **kwargs):
        super().__init__(**kwargs)

        self.bq_hook = get_bigquery_hook(bq_conn)
        self.query = query
        self.file = file

    def execute(self, context):
        # Grab phone_number column from BigQuery source table as a pandas dataframe
        df = self.bq_hook.get_pandas_df(self.query)

        # Check for directory in local FS. If it doesn't exist, create one.
        outdir = "/tmp/twilio"
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        # save dataframe to local system directory
        df.to_csv(os.path.join(outdir, self.file), index=False, header=False)


class CsvToTwilioToCsv(BaseOperator):
    # @apply_defaults
    def __init__(
        self, *, twilio_conn: str, source_file: str, target_file: str, **kwargs
    ):
        super().__init__(**kwargs)
        self.twilio_client = get_twilio_client(twilio_conn)
        self.source_file = source_file
        self.target_file = target_file

    def execute(self, context):
        with open(self.target_file, "a+") as target:
            with open(self.source_file, "r") as source:
                for num in source:
                    try:
                        self.twilio_client.lookups.v1.phone_numbers(num).fetch(
                            country_code="US"
                        )
                        target.write(f"{num.rstrip()},True\n")
                    except Exception:
                        target.write(f"{num.rstrip()},False\n")



from abc import ABC, abstractmethod, abstractproperty

from airflow.models.dag import DAG
from airflow.utils.task_group import TaskGroup

from gcp_airflow_foundations.base_class.data_source_table_config import (
    DataSourceTablesConfig,
)
from gcp_airflow_foundations.base_class.source_table_config import SourceTableConfig
from gcp_airflow_foundations.base_class.source_template_config import SourceTemplateConfig
from gcp_airflow_foundations.base_class.source_config import SourceConfig
from gcp_airflow_foundations.enums.schema_source_type import SchemaSourceType
from gcp_airflow_foundations.common.gcp.load_builder import load_builder
from gcp_airflow_foundations.common.utils.convert_config import convert_template_to_table
from gcp_airflow_foundations.source_class.schema_source_config import (
    AutoSchemaSourceConfig,
    GCSSchemaSourceConfig,
    BQLandingZoneSchemaSourceConfig,
)

import logging
import re


class DagBuilder(ABC):
    """A base DAG builder for creating a list of DAGs for a given source.
    Attributes
    ----------
    sources : list(DagBuilder)
        List of initialized subclasses of this class (dynamically updated during runtime)
    config : DataSourceTablesConfig
        DAG configuration object
    """

    sources = []

    def __init__(self, default_task_args: dict, config: DataSourceTablesConfig):
        self.default_task_args = default_task_args
        self.config = config
        self.validate_extra_options()
        self.set_schema_method_type()
        self.validate_custom_method()
        super().__init__()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "source_type"):
            raise AttributeError(
                f"{cls} source class is missing 'source_type' attribute"
            )
        cls.sources.append(cls)

    def build_dags(self):
        """
        Main DAG building method
        """
        data_source = self.config.source
        logging.info(f"Building DAG for {data_source.name}")

        dags = []
        for table_config in self.config.tables:
            dag = self.create_dag(table_config)
            dags.append(dag)

        # loop through templates
        for template_config in self.config.templates:
            table_list = self.get_templated_table_list(template_config.template_ingestion_options)

            if template_config.template_ingestion_options.dag_creation_mode == "TABLE":
                for table in table_list:
                    table_config = convert_template_to_table(template_config, table)
                    dag = self.create_dag(table_config)
                    dags.append(dag)
            else:
                dag = self.create_dag_source_level(template_config, table_list)
                dags.append(dag)

        extra_dags = self.get_extra_dags()
        if extra_dags is not None and not extra_dags == []:
            dags = dags + self.get_extra_dags()

        return dags

    @abstractmethod
    def get_bq_ingestion_task(self, table_config: SourceTableConfig):
        """Abstract method for the Airflow task that ingests data to the BigQuery staging table"""
        pass

    def get_datastore_ingestion_task(
        self,
        dag,
        preceding_task,
        data_source: SourceConfig,
        table_config: SourceTableConfig,
    ):
        """Method for the Airflow task group that upserts data to the ODS and HDS tables"""
        load_builder(
            data_source=data_source,
            table_config=table_config,
            schema_config=BQLandingZoneSchemaSourceConfig,
            preceding_task=preceding_task,
            dag=dag,
        )

    def get_extra_dags(self):
        # Override if non-table ingestion DAGs are needed for the flow: return them here
        return []

    @abstractmethod
    def set_schema_method_type(self):
        pass

    def get_schema_method_class(self):
        if self.schema_source_type == SchemaSourceType.GCS:
            return GCSSchemaSourceConfig

        elif self.schema_source_type == SchemaSourceType.AUTO:
            return AutoSchemaSourceConfig

        else:
            return

    def validate_custom_method(self):
        if self.schema_source_type == SchemaSourceType.CUSTOM:
            assert (
                self.get_schema_method_class() is not None
            ), "A custom method for schema parsing must be provided"

    @abstractmethod
    def validate_extra_options(self):
        pass

    def default_task_args_for_table(
        self, config: DataSourceTablesConfig, table_config: SourceTableConfig
    ):
        return {
            **self.default_task_args,
            "start_date": config.table_start_date(table_config),
        }

    def create_dag(self, table_config: SourceTableConfig):
        data_source = self.config.source
        table_default_task_args = self.default_task_args_for_table(
            self.config, table_config
        )

        logging.info(f"table_default_task_args {table_default_task_args}")

        kwargs = data_source.dag_args if data_source.dag_args else {}

        with DAG(
            dag_id=f"{data_source.name}.{table_config.table_name}",
            description=f"{data_source.name} to BigQuery load for {table_config.table_name}",
            schedule_interval=data_source.ingest_schedule,
            default_args=table_default_task_args,
            catchup=data_source.catchup,
            render_template_as_native_obj=True,
            **kwargs,
        ) as dag:

            with TaskGroup(group_id=table_config.table_name) as table_task_group:
                self.create_dag_tasks(dag, data_source, table_config)
            table_task_group
            return dag

    def create_dag_source_level(self, template_config: SourceTemplateConfig, table_names):
        data_source = self.config.source
        table_default_task_args = self.default_task_args_for_table(
            self.config, template_config
        )
        logging.info(f"table_default_task_args {table_default_task_args}")

        kwargs = data_source.dag_args if data_source.dag_args else {}

        with DAG(
            dag_id=f"{data_source.name}.{template_config.template_ingestion_options.ingestion_name}",
            description=f"{data_source.name} to BigQuery load for {template_config.template_ingestion_options.ingestion_name}",
            schedule_interval=data_source.ingest_schedule,
            default_args=table_default_task_args,
            catchup=data_source.catchup,
            render_template_as_native_obj=True,
            **kwargs,
        ) as dag:

            for i in range(len(table_names)):
                table_config = convert_template_to_table(template_config, table_names[i], i)
                with TaskGroup(group_id=table_names[i]) as table_task_group:
                    self.create_dag_tasks(dag, data_source, table_config)
                table_task_group

            return dag

    def create_dag_tasks(self, dag, data_source, table_config):
        load_to_bq_landing = self.get_bq_ingestion_task(dag, table_config)
        self.get_datastore_ingestion_task(
            dag, load_to_bq_landing, data_source, table_config
        )

    def get_templated_table_list(self, templated_ingestion_options):
        if templated_ingestion_options.ingest_mode.value == "INGEST_BY_TABLE_NAMES":
            return templated_ingestion_options.table_names

        full_table_list = self.get_source_tables_to_ingest()

        if templated_ingestion_options.ingest_mode.value == "INGEST_ALL":
            return full_table_list

        regex_pattern = templated_ingestion_options.regex_pattern
        r = re.compile(regex_pattern)
        return list(filter(r.match, full_table_list))

    def get_source_tables_to_ingest(self):
        return []



import json
import logging

from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook


from airflow.models.baseoperator import BaseOperator
# from airflow.utils.decorators import apply_defaults

from gcp_airflow_foundations.operators.api.hooks.twilio_hook import TwilioHook
from urllib.parse import urlparse
from airflow.providers.google.cloud.hooks.gcs import GCSHook


class TwilioToBigQueryOperator(BaseOperator):
    """
    Twilio Text Message data to BigQuery

    Attributes:
        twilio_account_sid: Twilio Account SID to be used in the API
        project_id: GCP project ID
        table_id: Staging table name
        dataset_id: Staging dataset name
        schema_fields: Schema fields provided by the Airflow Framework for the Twilio Staging table
        delegate_to: The account to impersonate using domain-wide delegation of authority, if any
        gcp_conn_id: Airflow GCP connection ID
        twilio_conn_id: Airflow Twilio connection ID
        column_mapping: Column mapping dictionary
    """

    template_fields = ("dataset_id", "table_id", "project_id", "labels")

    # pylint: disable=too-many-arguments
    # @apply_defaults
    def __init__(
        self,
        twilio_account_sid,
        dataset_id,
        table_id,
        project_id=None,
        schema_fields=None,
        gcs_schema_object=None,
        time_partitioning=None,
        gcp_conn_id="google_cloud_default",
        twilio_conn_id="twilio_default",
        delegate_to=None,
        labels=None,
        encryption_configuration=None,
        *args,
        **kwargs
    ):

        super(TwilioToBigQueryOperator, self).__init__(*args, **kwargs)

        self.twilio_account_sid = twilio_account_sid
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.schema_fields = schema_fields
        self.gcs_schema_object = gcs_schema_object
        self.gcp_conn_id = gcp_conn_id
        self.twilio_conn_id = twilio_conn_id
        self.delegate_to = delegate_to
        self.time_partitioning = {} if time_partitioning is None else time_partitioning
        self.labels = labels
        self.encryption_configuration = encryption_configuration

    def execute(self, context):
        twilio_hook = TwilioHook(twilio_conn_id=self.twilio_conn_id)

        bq_hook = BigQueryHook(gcp_conn_id=self.gcp_conn_id)
        # conn = bq_hook.get_conn()
        # cursor = conn.cursor()

        if not bq_hook.table_exists(
            project_id=self.project_id,
            dataset_id=self.dataset_id,
            table_id=self.table_id,
        ):

            if not self.schema_fields and self.gcs_schema_object:

                parsed_url = urlparse(self.gcs_schema_object)
                gcs_bucket = parsed_url.netloc
                gcs_object = parsed_url.path.lstrip("/")

                gcs_hook = GCSHook(
                    gcp_conn_id=self.google_cloud_storage_conn_id,
                    delegate_to=self.delegate_to,
                )
                schema_fields = json.loads(
                    gcs_hook.download(gcs_bucket, gcs_object).decode("utf-8")
                )
            else:
                schema_fields = self.schema_fields

            # cursor.create_empty_table(
            bq_hook.create_empty_table(
                project_id=self.project_id,
                dataset_id=self.dataset_id,
                table_id=self.table_id,
                schema_fields=schema_fields,
                time_partitioning=self.time_partitioning,
                labels=self.labels,
                encryption_configuration=self.encryption_configuration,
            )

        service = bq_hook.get_service()

        body = {
            "rows": twilio_hook.get_text_messages(account_sid=self.twilio_account_sid)
        }

        service_exec = service.tabledata().insertAll(
            body=body,
            projectId=self.project_id,
            tableId=self.table_id,
            datasetId=self.dataset_id,
        )

        results = service_exec.execute()

        logging.info(results)



from google.cloud.bigquery.table import TableReference


# https://cloud.google.com/dlp/docs/reference/rest/v2/projects.jobTriggers#JobTrigger
def build_job_trigger(
    display_name: str,
    template_name: str,
    table_ref: TableReference,
    rows_limit_percent: int,
    results_table_ref: TableReference,
    recurrence: str,
):
    schedule = {"recurrence_period_duration": {"seconds": 30 * 60 * 60 * 24}}
    return {
        "display_name": display_name,
        "description": "DLP scan for {}.{} triggered by Airflow".format(
            table_ref.dataset_id, table_ref.table_id
        ),
        "inspect_job": build_inspect_job_config(
            template_name, table_ref, rows_limit_percent, results_table_ref
        ),
        "triggers": [{"schedule": schedule}],
        "status": "HEALTHY",
    }


def build_inspect_job_config(
    template_name, table_ref, rows_limit_percent, results_table_ref
):
    """ Create an inspection job config: https://cloud.google.com/dlp/docs/creating-job-triggers"""
    return {
        "storage_config": __bq_storage_inspect_job_config(
            table_ref, rows_limit_percent
        ),
        "inspect_template_name": template_name,
        "actions": [__action_save_to_bq(results_table_ref)],
    }


def __bq_storage_inspect_job_config(table_ref, rows_limit_percent):
    return {
        "big_query_options": {
            "table_reference": {
                "project_id": table_ref.project,
                "dataset_id": table_ref.dataset_id,
                "table_id": table_ref.table_id,
            },
            "rows_limit_percent": rows_limit_percent,
            "sample_method": "RANDOM_START",
        }
    }


def __action_save_to_bq(table_ref):
    return {
        "save_findings": {
            "output_config": {
                "output_schema": "BIG_QUERY_COLUMNS",
                "table": {
                    "project_id": table_ref.project,
                    "dataset_id": table_ref.dataset_id,
                    "table_id": table_ref.table_id,
                },
            }
        }
    }



import logging


def get_twilio_schema(tableId=None):
    if tableId == "twilio_text_messages":
        schema_fields = [
            {"name": "account_sid", "type": "STRING", "mode": "NULLABLE"},
            {"name": "api_version", "type": "STRING", "mode": "NULLABLE"},
            {"name": "body", "type": "STRING", "mode": "NULLABLE"},
            {"name": "date_created", "type": "TIMESTAMP", "mode": "NULLABLE"},
            {"name": "date_sent", "type": "TIMESTAMP", "mode": "NULLABLE"},
            {"name": "date_updated", "type": "TIMESTAMP", "mode": "NULLABLE"},
            {"name": "direction", "type": "STRING", "mode": "NULLABLE"},
            {"name": "error_code", "type": "STRING", "mode": "NULLABLE"},
            {"name": "error_message", "type": "STRING", "mode": "NULLABLE"},
            {"name": "from", "type": "STRING", "mode": "NULLABLE"},
            {"name": "messaging_service_sid", "type": "STRING", "mode": "NULLABLE"},
            {"name": "num_media", "type": "INTEGER", "mode": "NULLABLE"},
            {"name": "num_segments", "type": "INTEGER", "mode": "NULLABLE"},
            {"name": "price", "type": "FLOAT", "mode": "NULLABLE"},
            {"name": "price_unit", "type": "STRING", "mode": "NULLABLE"},
            {"name": "sid", "type": "STRING", "mode": "NULLABLE"},
            {"name": "status", "type": "STRING", "mode": "NULLABLE"},
            {"name": "to", "type": "STRING", "mode": "NULLABLE"},
            {"name": "uri", "type": "STRING", "mode": "NULLABLE"},
            {"name": "feedback", "type": "STRING", "mode": "NULLABLE"},
            {"name": "media", "type": "STRING", "mode": "NULLABLE"},
        ]
        return schema_fields
    else:
        logging.info(f"Schema unavailable for {tableId}")
        raise Exception(f"Schema unavailable for {tableId}")



from gcp_airflow_foundations.base_class.source_table_config import SourceTableConfig
from gcp_airflow_foundations.base_class.source_template_config import SourceTemplateConfig
import logging


def convert_template_to_table(template_config: SourceTemplateConfig, table_name: str, i: int):

    landing_zone_table_name_override = \
        template_config.landing_zone_table_name_override_template.replace("{table}", table_name)
    dest_table_override = \
        template_config.dest_table_override_template.replace("{table}", table_name)

    if template_config.surrogate_keys:
        surrogate_keys = template_config.surrogate_keys["table"]
    else:
        surrogate_keys = []

    params = {
        "table_name": table_name,
        "ingestion_type": template_config.ingestion_type,
        "landing_zone_table_name_override": landing_zone_table_name_override,
        "dest_table_override": dest_table_override,
        "surrogate_keys": surrogate_keys,
        "start_date_tz": template_config.start_date_tz,
        "version": template_config.version,
        "catchup": template_config.catchup,
    }

    optional_fields = [
        "dest_table_override",
        "surrogate_keys",
        "column_mapping",
        "cluster_fields",
        "column_casting",
        "new_column_udfs",
        "ods_config",
        "hds_config",
        "start_date",
    ]

    for of in optional_fields:
        if hasattr(template_config, of):
            params[of] = getattr(template_config, of)

    params["extra_options"] = convert_template_extra_options(template_config, i)

    return SourceTableConfig(**params)


def convert_template_extra_options(template_config: SourceTemplateConfig, i: int):
    """
    Converts the extra_options field in a SourceTemplateConfig to match a corresponding SourceTableConfig.
    """
    template_extra_options = getattr(template_config, "extra_options")
    table_extra_options = {}
    for config_option_name, _ in template_extra_options.items():
        # get each second-level dict in extra_options corresponding to a particular source specific config
        config_options = template_extra_options[config_option_name]
        table_config_options = {}
        # loop through the config options in that config
        for key, val in config_options.items():
            if key in template_config.iterable_options:
                # if the config name matches to a value in iterable_options, parse as a list
                table_config_options[key] = val[i]
            else:
                # else, value applies to all tables
                table_config_options[key] = val
        table_extra_options[config_option_name] = table_config_options
    return table_extra_options



from airflow.exceptions import AirflowException

from dacite import Config
from dataclasses import dataclass, field
from dacite import from_dict


from pydantic import validator, root_validator

import datetime
import re
from typing import List, Optional

from regex import E

from gcp_airflow_foundations.base_class.ods_metadata_config import (
    OdsTableMetadataConfig,
)
from gcp_airflow_foundations.enums.ingestion_type import IngestionType
from gcp_airflow_foundations.enums.template_ingestion import TemplateIngestionMode

from gcp_airflow_foundations.base_class.source_base_config import SourceBaseConfig
from gcp_airflow_foundations.base_class.ods_table_config import OdsTableConfig
from gcp_airflow_foundations.base_class.hds_table_config import HdsTableConfig
from gcp_airflow_foundations.base_class.template_ingestion_options_config import TemplateIngestionOptionsConfig


@dataclass
class SourceTemplateConfig(SourceBaseConfig):
    """
    Template configuration data class. Template configuration applies to all tables specified within it, either by
    explicitly providing a list of table_names, or by fetching a full list of source tables and optionally filtering
    by regex expression. Templates can create either one DAG per table or one DAG per template.

    Attributes:
        surrogate_keys : Keys used to identify unique records when merging into ODS.
        column_mapping : Mapping used to rename columns.
        cluster_fields: The fields used for clustering. BigQuery supports clustering for both partitioned and non-partitioned tables.
        landing_zone_table_name_override_template: Optional staging zone table name template.
        dest_table_override_template: Optional target table name template. If None, uses table_name instead.
        column_casting : Mapping used to cast columns into a specific data type. Note column name uses that of the landing zone table.
        ods_config : ODS table configuration. See :class:`gcp_airflow_foundations.base_class.ods_table_config.OdsTableConfig`.
        hds_config : HDS table confidwguration. See :class:`gcp_airflow_foundations.base_class.hds_table_config.HdsTableConfig`.
        iterable_options: list of config options within the source extra_options to iterate through for DAG creation.
        template_ingestion_options: Configuration for template-level ingestion.
        extra_options: Field for storing additional configuration options.
        start_date : Start date override for DAG
        start_date_tz : Timezone
        version : The Dag version for the table. Can be incremented if logic changes.
        catchup : Passed to a dag [see doc](https://airflow.apache.org/docs/apache-airflow/stable/dag-run.html#catchup).
            Defaults to True. May want to change it to False if Dag version is changed, and we don't want to rerun past dags.
    """
    start_date: Optional[str]
    ingestion_type: IngestionType  # FULL or INCREMENTAL
    surrogate_keys: Optional[dict]
    column_mapping: Optional[dict]
    cluster_fields: Optional[List[str]]
    column_casting: Optional[dict]
    new_column_udfs: Optional[dict]
    hds_config: Optional[HdsTableConfig]
    template_ingestion_options: TemplateIngestionOptionsConfig
    extra_options: dict = field(default_factory=dict)
    iterable_options: list = field(default_factory=list)
    dest_table_override_template: Optional[str] = "{table}"
    landing_zone_table_name_override_template: Optional[str] = "{table}"
    start_date_tz: Optional[str] = "EST"
    ods_config: Optional[OdsTableConfig] = OdsTableConfig(
        ods_metadata=OdsTableMetadataConfig(),
        ods_table_time_partitioning=None,
        partition_column_name=None,
    )
    version: int = 1
    catchup: bool = True

    # Override values
    def __post_init__(self):
        if self.extra_options is None:
            self.extra_options = {}

        if self.landing_zone_table_name_override_template is None:
            self.landing_zone_table_name_override_template = "{table}"

        if self.dest_table_override_template is None:
            self.dest_table_override_template = "{table}"

        if self.iterable_options is None:
            self.iterable_options = []

    @root_validator(pre=True)
    def valid_template_ingestion_options(cls, values):
        if "template_ingestion_options" in values:
            options = values["template_ingestion_options"]
            template_ingestion = options["ingest_mode"]

            if template_ingestion.value == TemplateIngestionMode.INGEST_BY_TABLE_NAMES:
                assert (
                    not options["regex_pattern"]
                ), "If table_names are explicitly provided for a template, regex_pattern should not be provided in template_ingestion_options"
            elif template_ingestion.value == TemplateIngestionMode.INGEST_ALL:
                assert (
                    not options["regex_pattern"] and not options["table_names"]
                ), "If a template is ingesting all tables, regex_pattern and table_names should not be provided in template_ingestion_options"

            elif template_ingestion.value == TemplateIngestionMode.INGEST_BY_REGEX:
                assert (
                    re.compile(options["regex_pattern"])
                ), "If ingest_mode is set to 'INGEST_BY_REGEX', the regex_pattern should be a valid regex pattern"

            return values

    @root_validator(pre=True)
    def valid_iterable_options(cls, values):
        if "iterable_options" in values:
            iterable_options = values["iterable_options"]
            if iterable_options == []:
                table_names = values["template_ingestion_options"]["table_names"]
                assert (
                    not table_names == []
                ), "table_names should be explicitly provided for a template if using iterable_options"

            return values



from dataclasses import dataclass, field
from typing import Optional

from gcp_airflow_foundations.base_class.dlp_source_config import DlpSourceConfig


@dataclass
class DlpTableConfig:
    """
    Attributes:
    """

    is_on: Optional[bool] = None
    template_name_override: Optional[str] = None
    rows_limit_percent_override: Optional[int] = None
    min_match_count_override: Optional[int] = None
    recurrence_override: Optional[str] = None

    def set_source_config(self, source_config: DlpSourceConfig):
        self.source_config = source_config

    def get_is_on(self):
        return self.is_on if self.is_on else (self.source_config is not None)

    def get_template_name(self):
        return (
            self.template_name_override
            if self.template_name_override
            else self.source_config.template_name
        )

    def get_rows_limit_percent(self):
        return (
            self.rows_limit_percent_override
            if self.rows_limit_percent_override
            else self.source_config.rows_limit_percent
        )

    def get_min_match_count(self):
        return (
            self.min_match_count_override
            if self.min_match_count_override
            else self.source_config.min_match_count
        )

    def get_recurrence(self):
        return (
            self.recurrence_override
            if self.recurrence_override
            else self.source_config.recurrence_schedule
        )



from airflow.hooks.base import BaseHook
from typing import Optional
from airflow.exceptions import AirflowException

import requests
import json


class DataformHook(BaseHook):
    '''
    Airflow Hook to connect to Dataform's API and run Dataform jobs.
    This hook will POST to Dataform's ApiService_RunCreate to initiate a new dataform run.
    If no tags are provided, Dataform will run all.

    Attributes:
    :param project_id: project_id
    :type project_id: str
    :param environment: name of the environment, e.g. production, development...
    :type environment: str
    :param schedule: the name of the schedule that will be triggered by airflow
    :type schedule: str
    :param tags: (optional) A list of tags with which the action must have in order to be run. If not set, then all actions will run.
    :type tags: list[str]

    Instructions to prepare Dataform for the API call:
    1. create schedule in Dataform for REST API call
        a) environments.json >> under the appropriate environment (e.g. prod, dev..) click on "create new schedule"
        b) enter schedule name (e.g. composer)
        c) disable "Enable this schedule"
    2. get Dataform project ID
        easiest way is to look at the url
        E.g. https://app.dataform.co/#/<project_id>/b/<branch>/file/<file-name>
    3. generate API token:
        project settings >> API keys >> GENERATE NEW API KEY



    - Dataform documentation on using REST API: https://docs.dataform.co/dataform-web/api
    - Helpful Medium tutorial: https://medium.com/google-cloud/cloud-composer-apache-airflow-dataform-bigquery-de6e3eaabeb3

    ** Note: Schedules must be on the master branch. In Dataform, you'll have to create a branch first and then merge changes into master.
    '''

    def __init__(self, dataform_conn_id='datafrom_default') -> None:
        self.conn = self.get_connection(dataform_conn_id)
        self.api_key = self.conn.password
        self.headers = {'Authorization': f'Bearer {self.api_key}'}

    def run_job(self, project_id: str, environment: str, schedule: str, tags: Optional[str] = []) -> str:

        base_url = f'https://api.dataform.co/v1/project/{project_id}/run'
        # headers = {'Authorization': f'Bearer {self.api_key}'}
        run_create_request = {
            "environmentName": environment,
            "scheduleName": schedule,
            "runConfig": {
                "tags": tags
            }
        }

        # ApiService_RunCreate: Initiates new dataform runs within the project, and returns the ID of any created runs.
        response = requests.post(base_url, data=json.dumps(run_create_request), headers=self.headers)

        try:
            # ApiService_RunGet: Returns information about a specific run
            run_url = base_url + '/' + response.json()['id']
            return run_url
        except Exception:
            raise AirflowException(f"Dataform ApiService Error: {response.json()}")

    def check_job_status(self, run_url: str) -> str:
        response = requests.get(run_url, headers=self.headers)
        return response.json()['status']



from airflow.exceptions import AirflowException

from dacite import Config
from dataclasses import dataclass, field
from dacite import from_dict


from pydantic import validator, root_validator

import datetime
from typing import List, Optional

from gcp_airflow_foundations.base_class.ods_metadata_config import (
    OdsTableMetadataConfig,
)
from gcp_airflow_foundations.enums.ingestion_type import IngestionType
from gcp_airflow_foundations.enums.hds_table_type import HdsTableType

from gcp_airflow_foundations.base_class.source_base_config import SourceBaseConfig
from gcp_airflow_foundations.base_class.ods_table_config import OdsTableConfig
from gcp_airflow_foundations.base_class.hds_table_config import HdsTableConfig
from gcp_airflow_foundations.base_class.facebook_table_config import FacebookTableConfig
from gcp_airflow_foundations.enums.facebook import ApiObject
from gcp_airflow_foundations.base_class.column_udf_config import ColumnUDFConfig


@dataclass
class SourceTableConfig(SourceBaseConfig):
    """
    Table configuration data class.

    Attributes:
        table_name : Table name. Used for Dag Id.
        ingestion_type: FULL or INCREMENTAL.
        landing_zone_table_name_override: Optional staging zone table name.
        dest_table_override: Optional target table name. If None, use table_name instead.
        surrogate_keys : Keys used to identify unique records when merging into ODS.
        column_mapping : Mapping used to rename columns.
        cluster_fields: The fields used for clustering. BigQuery supports clustering for both partitioned and non-partitioned tables.
        column_casting : Mapping used to cast columns into a specific data type. Note column name uses that of the landing zone table.
        ods_config : ODS table configuration. See :class:`gcp_airflow_foundations.base_class.ods_table_config.OdsTableConfig`.
        hds_config : HDS table configuration. See :class:`gcp_airflow_foundations.base_class.hds_table_config.HdsTableConfig`.
        facebook_table_config: Extra options for ingesting data from the Facebook API.
        extra_options: Field for storing additional configuration options.
        start_date : Start date override for DAG
        start_date_tz : Timezone
        version : The Dag version for the table. Can be incremented if logic changes.
        catchup : Passed to a dag [see doc](https://airflow.apache.org/docs/apache-airflow/stable/dag-run.html#catchup).
            Defaults to True. May want to change it to False if Dag version is changed, and we don't want to rerun past dags.
    """

    table_name: str
    ingestion_type: IngestionType  # FULL or INCREMENTAL
    landing_zone_table_name_override: Optional[str]
    dest_table_override: Optional[str]
    surrogate_keys: List[str]
    column_mapping: Optional[dict]
    cluster_fields: Optional[List[str]]
    column_casting: Optional[dict]
    new_column_udfs: Optional[dict]
    hds_config: Optional[HdsTableConfig]
    start_date: Optional[str]
    extra_options: dict = field(default_factory=dict)
    facebook_table_config: Optional[FacebookTableConfig] = FacebookTableConfig(
        api_object=ApiObject("INSIGHTS"),
        breakdowns=[],
        action_breakdowns=[]
    )
    start_date_tz: Optional[str] = "EST"
    ods_config: Optional[OdsTableConfig] = OdsTableConfig(
        ods_metadata=OdsTableMetadataConfig(),
        ods_table_time_partitioning=None,
        partition_column_name=None,
    )
    version: int = 1
    catchup: bool = True

    # Override values
    def __post_init__(self):
        if self.landing_zone_table_name_override is None:
            self.landing_zone_table_name_override = self.table_name

        if self.extra_options is None:
            self.extra_options = {}



from dataclasses import fields
from os import X_OK
from urllib.parse import urlparse
from dacite import from_dict
from dataclasses import dataclass

from airflow.operators.dummy import DummyOperator
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.operators.python import PythonOperator
from airflow.providers.google.cloud.sensors.gcs import GCSObjectExistenceSensor

from gcp_airflow_foundations.operators.api.sensors.gcs_sensor import GCSObjectListExistenceSensor
from gcp_airflow_foundations.operators.api.sensors.gcs_prefix_sensor import GCSObjectPrefixListExistenceSensor
from gcp_airflow_foundations.source_class.generic_file_source import GenericFileIngestionDagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder


class GCSFileIngestionDagBuilder(GenericFileIngestionDagBuilder):
    """
    Builds DAGs to load files from GCS to a BigQuery Table.

    For GCS->BQ ingestion, either a metadata file is required or the field templated_file_name must be provided.
    If a metadata file is provided, itt can be a fixed file, or can be a new file supplied daily.
    Airflow context variables are supported for the file naming, e.g.
        TABLE_METADATA_FILE_{{ ds }}.csv
    for a metadata file supplied daily.

    The format of the metadata file should be a csv with one column as follows:
        FILE_NAME_1
        ...
        FILE_NAME_N
    with all files to ingest
    """
    source_type = "GCS"

    def flag_file_sensor(self, table_config, taskgroup):
        if "flag_file_path" in table_config.extra_options.get("file_table_config"):
            flag_file_path = table_config.extra_options.get("file_table_config")["flag_file_path"]
            bucket = self.config.source.extra_options["gcs_bucket"]
            return GCSObjectExistenceSensor(
                task_id="wait_for_flag_file",
                bucket=bucket,
                object=flag_file_path,
                task_group=taskgroup
            )
        else:
            return None

    def file_ingestion_task(self, table_config, taskgroup):
        """
        No ingestion is needed - data is already in GCS, so return a dummy operator.
        """
        return None

    def file_sensor(self, table_config, taskgroup):
        """
        Returns an Airflow sensor that waits for the list of files specified by the metadata file provided.
        """
        bucket = self.config.source.extra_options["gcs_bucket"]
        table_name = table_config.table_name
        files_to_wait_for = "{{ ti.xcom_pull(key='file_list', task_ids='" + table_name + ".ftp_taskgroup.get_file_list') }}"

        if self.config.source.extra_options["file_source_config"]["file_prefix_filtering"]:
            return GCSObjectPrefixListExistenceSensor(
                task_id="wait_for_files_to_ingest",
                bucket=bucket,
                prefixes=files_to_wait_for,
                task_group=taskgroup
            )
        else:
            return GCSObjectListExistenceSensor(
                task_id="wait_for_files_to_ingest",
                bucket=bucket,
                objects=files_to_wait_for,
                task_group=taskgroup
            )

    def delete_files(self, table_config, **kwargs):
        ti = kwargs["ti"]
        table_name = table_config.table_name
        files_to_load = ti.xcom_pull(key='loaded_files', task_ids=f'{table_name}.ftp_taskgroup.load_gcs_to_landing_zone')
        data_source = self.config.source
        bucket = data_source.extra_options["gcs_bucket"]
        gcs_hook = GCSHook()

        for file in files_to_load:
            gcs_hook.delete(bucket_name=bucket, object_name=file)

    def delete_gcs_files(self, table_config, taskgroup):
        return PythonOperator(
            task_id="delete_gcs_files",
            op_kwargs={"table_config": table_config},
            python_callable=self.delete_files,
            task_group=taskgroup
        )

    def validate_extra_options(self):
        # GCS Source only requires the checks for the base file_source_config and file_table_configs:
        # other sources like SFTP require extra checks
        super().validate_extra_options()



import pandas as pd
import json
import pyarrow.parquet as pq
import pyarrow

from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
# from airflow.utils.decorators import apply_defaults

from google.cloud import bigquery


# @apply_defaults
class CustomBigQueryHook(BigQueryHook):
    """
    Extends the default BigQuery Hook with custom methods to interact with Google Cloud BigQuery.

    :param gcp_conn_id: The Airflow connection used for GCP credentials.
    :type gcp_conn_id: Optional[str]
    """

    def __init__(self, gcp_conn_id: str, **kwargs) -> None:
        super().__init__(gcp_conn_id=gcp_conn_id, **kwargs)

    def load_nested_table_from_dataframe(
        self,
        df: pd.DataFrame,
        project_id: str,
        destination_dataset_table: str,
        write_disposition: str = "WRITE_TRUNCATE",
    ) -> bigquery.job.LoadJob:
        """
        Loads a pandas.DataFrame to a BigQuery table. It supports nested fields.

        Returns a BigQuety LoadJob object.

        :param df: The Dataframe to load in BigQuery
        :type df: pandas.DataFrame
        :param project_id: The Google Cloud Project ID
        :type project_id: str
        :param destination_dataset_table: The full table name to load the data. Format: (<project>.|<project>:)<dataset>.<table>
        :type destination_dataset_table: str
        :param write_disposition: Specifies the action that occurs if the destination table already exists. (default: 'WRITE_TRUNCATE')
        :type write_disposition: str
        """

        client = self.get_client(project_id=project_id)

        writer = pyarrow.BufferOutputStream()
        pq.write_table(
            pyarrow.Table.from_pandas(df), writer, use_compliant_nested_type=True
        )
        reader = pyarrow.BufferReader(writer.getvalue())

        parquet_options = bigquery.format_options.ParquetOptions()
        parquet_options.enable_list_inference = True

        job_config = bigquery.LoadJobConfig()
        job_config.source_format = bigquery.SourceFormat.PARQUET
        job_config.parquet_options = parquet_options
        job_config.write_disposition = write_disposition

        return client.load_table_from_file(
            reader, destination_dataset_table, job_config=job_config
        )



from dataclasses import dataclass, field
from pydantic import validator, root_validator


@dataclass
class PolicyTagConfig:
    """
    Information about a policy tag to apply the columns
    See the following for a good example of creating a taxonomy: https://cloud.google.com/bigquery/docs/best-practices-policy-tags
    Attributes:
         taxonomy: The taxonomy to use: https://cloud.google.com/data-catalog/docs/samples/data-catalog-ptm-create-taxonomy
         location: GCP logcation where the taxonomy was created
         tag: The tag to use (must be part of the taxonomy)
    """

    taxonomy: str
    location: str
    tag: str


@dataclass
class DlpSourceConfig:
    """
    Attributes:
         results_dataset_id:  Dataset for temporarily storing dlp results
         template_name: Template to use when running DLP. Full path is expected: projects/{prj_id}/locations/{location}}/inspectTemplates/{template}
         rows_limit_percent: What percentage of rows to scan/sample
         policy_tag_config: Policy tag to use for tagging columns
         min_match_count: Min numbers of rows that must match and InfoType for the column to be tagged
         recurrence_schedule: How often to run the DLP scan and update the tags
         run_on_first_execution: True if we should run on the first time we ingest a new table
    """

    results_dataset_id: str
    template_name: str
    policy_tag_config: PolicyTagConfig
    rows_limit_percent: int = 10
    min_match_count: int = 1
    recurrence_schedule: str = "0 0 1 * *"  # First day of every month
    run_on_first_execution: bool = True



from dacite import Config
from dataclasses import dataclass, field
from pydantic import validator
from typing import List, Optional

from gcp_airflow_foundations.enums.facebook import ApiObject


@dataclass
class FacebookTableConfig:
    """
    Attributes:
        api_object: The API object to query {insights, campaign, adset}
        breakdowns: How to break down the result. For more than one breakdown, only certain combinations are available.
        action_breakdowns: How to break down action results. Supports more than one breakdowns. Default value is ["action_type"].
    """

    api_object: Optional[ApiObject]
    breakdowns: Optional[List[str]]
    action_breakdowns: Optional[List[str]]

    @validator("api_object")
    def valid_fields(cls, v):
        if v is None:
            return ApiObject.INSIGHTS
        else:
            return v



from typing import Optional, List

# from airflow.utils.decorators import apply_defaults

from airflow.exceptions import AirflowException

import logging

from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryCreateEmptyTableOperator,
)


class CustomBigQueryCreateEmptyTableOperator(BigQueryCreateEmptyTableOperator):
    # @apply_defaults
    def __init__(
        self,
        *,
        project_id: str,
        dataset_id: str,
        table_id: str,
        dag_table_id: str,
        cluster_fields: Optional[List[str]] = None,
        time_partitioning: Optional[dict] = None,
        gcp_conn_id: str = "google_cloud_default",
        **kwargs,
    ) -> None:
        super(CustomBigQueryCreateEmptyTableOperator, self).__init__(
            project_id=project_id,
            dataset_id=dataset_id,
            table_id=table_id,
            table_resource=None,
            exists_ok=True,
            **kwargs,
        )

        self.table = table_id
        self.dataset_id = dataset_id
        self.time_partitioning = time_partitioning
        self.cluster_fields = cluster_fields
        self.schema_task_id = f"{dag_table_id}.schema_parsing"

    def pre_execute(self, context) -> None:
        logging.info(self.schema_task_id)
        schema_fields = self.xcom_pull(context=context, task_ids=self.schema_task_id)[
            f"{self.dataset_id}.{self.table_id}"
        ]

        if self.cluster_fields:
            self.table_resource = {
                "schema": {"fields": schema_fields},
                "timePartitioning": self.time_partitioning,
                "encryptionConfiguration": None,
                "labels": None,
                "clustering": {"fields": self.cluster_fields},
            }
        else:
            self.table_resource = {
                "schema": {"fields": schema_fields},
                "timePartitioning": self.time_partitioning,
                "encryptionConfiguration": None,
                "labels": None,
            }



import requests
import time
from datetime import datetime
import json
from requests.auth import HTTPBasicAuth

import hashlib

from airflow.hooks.base import BaseHook


class TwilioHook(BaseHook):
    """
    Airflow Hook to connect to Twilio's API and retrieve Text Message data
    """

    def __init__(self, twilio_conn_id="twilio_default") -> None:
        self.twilio_conn_id = (twilio_conn_id,)
        self.base_uri = "https://api.twilio.com/2010-04-01"

        connection = self.get_connection(twilio_conn_id)
        self.sid = connection.login
        self.secret = connection.password

    def transform_text_messages(self, text_messages) -> list:

        rows = []
        for i in text_messages:
            i["feedback"] = i["subresource_uris"]["feedback"]
            i["media"] = i["subresource_uris"]["media"]
            i.pop("subresource_uris", None)

            i["date_created"] = datetime.strptime(
                i["date_created"], "%a, %d %b %Y %H:%M:%S %z"
            ).strftime("%Y-%m-%d %H:%M:%S")
            i["date_sent"] = datetime.strptime(
                i["date_sent"], "%a, %d %b %Y %H:%M:%S %z"
            ).strftime("%Y-%m-%d %H:%M:%S")
            i["date_updated"] = datetime.strptime(
                i["date_updated"], "%a, %d %b %Y %H:%M:%S %z"
            ).strftime("%Y-%m-%d %H:%M:%S")

            keys = ["sid"]

            joined_keys = "".join([i[key] for key in keys])

            rows.append(
                {
                    "insertId": hashlib.md5(joined_keys.encode("utf-8")).hexdigest(),
                    "json": i,
                }
            )

        return rows

    def get_text_messages(self, account_sid, FromDate=None, ToDate=None) -> list:

        URL = f"{self.base_uri}/Accounts/{account_sid}/Messages.json"

        auth = HTTPBasicAuth(self.sid, self.secret)

        params = {"PageSize": 1000}

        if FromDate:
            params["DateSent>"] = FromDate

        if ToDate:
            params["DateSent<"] = ToDate

        text_messages = []

        while True:
            time.sleep(1)

            r = requests.get(URL, params=params, auth=auth)
            r.raise_for_status()

            body = r.json()
            text_messages.extend(body["messages"])

            NextPageURI = body.get("next_page_uri", None)
            if not NextPageURI:
                break

            URL = f"https://api.twilio.com/{NextPageURI}"

        rows = self.transform_text_messages(text_messages)

        return rows



from typing import Optional

from airflow.models import BaseOperator, BaseOperatorLink
from airflow.contrib.operators.bigquery_operator import (
    BigQueryOperator,
    BigQueryCreateEmptyTableOperator,
)

from airflow.utils.decorators import apply_defaults
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook

from airflow.exceptions import AirflowException

import logging

from gcp_airflow_foundations.operators.gcp.schema_migration.schema_migration_audit import (
    SchemaMigrationAudit,
)


class MigrateSchema(BaseOperator):
    """
    Detects any changes in the source table's schema and updates the target table's schema.

    :param project_id: GCP project ID
    :type project_id: str
    :param table_id: Target table name
    :type table_id: str
    :param dataset_id: Target dataset name
    :type dataset_id: str
    :param new_schema_fields: The current schema fields of the source table
    :type new_schema_fields: list
    :param delegate_to: The account to impersonate using domain-wide delegation of authority, if any
    :type delegate_to: str
    :param gcp_conn_id: Airflow GCP connection ID
    :type gcp_conn_id: str
    """

    @apply_defaults
    def __init__(
        self,
        *,
        dataset_id,
        table_id,
        dag_table_id,
        project_id,
        new_schema_fields=None,
        gcp_conn_id="google_cloud_default",
        delegate_to=None,
        encryption_configuration=None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.dag_table_id = dag_table_id
        self.new_schema_fields = new_schema_fields
        self.gcp_conn_id = gcp_conn_id
        self.delegate_to = delegate_to
        self.encryption_configuration = encryption_configuration

        self.hook = BigQueryHook(
            gcp_conn_id=self.gcp_conn_id, delegate_to=self.delegate_to
        )
        conn = self.hook.get_conn()
        self.cursor = conn.cursor()

    def execute(self, context):
        if not self.new_schema_fields:
            self.new_schema_fields = self.xcom_pull(
                context=context, task_ids=f"{self.dag_table_id}.schema_parsing"
            )[f"{self.dataset_id}.{self.table_id}"]

        (
            query,
            schema_fields_updates,
            sql_columns,
            change_log,
        ) = self.build_schema_query()

        if change_log:
            logging.info("Migrating new schema to target table")

            if sql_columns:
                self.hook.run_query(
                    sql=query,
                    use_legacy_sql=False,
                    destination_dataset_table=f"{self.project_id}.{self.dataset_id}.{self.table_id}",
                    write_disposition="WRITE_TRUNCATE",
                )

            if schema_fields_updates:
                self.hook.update_table_schema(
                    project_id=self.project_id,
                    dataset_id=self.dataset_id,
                    table_id=self.table_id,
                    schema_fields_updates=schema_fields_updates,
                    include_policy_tags=False,
                )

            SchemaMigrationAudit(
                project_id=self.project_id, dataset_id=self.dataset_id
            ).insert_change_log_rows(change_log)

        else:
            logging.info("No schema changes detected")

    def build_schema_query(self):
        """
        Schema change types supported:
        1) Column type
        2) New column
        3) Deleted column

        :returns: list of
                    string of the SQL query that will be executed in BigQuery to modify the table's schema,
                    list of the columns whose mode will be relaxed to nullable after being deleted in the source table. The list is used in the update_table_schema method of the BigQuery hook,
                    list of updated columns that is used to check whether there are any columns to be updated before proceeding to execute the SQL query,
                    list of rows of changes that is inserted in the schema migration audit table
        :rtype: list
        """

        self.current_schema_fields = self.hook.get_schema(
            project_id=self.project_id, dataset_id=self.dataset_id, table_id=self.table_id
        ).get("fields", None)

        logging.info(f"The current schema is: {self.current_schema_fields}")

        logging.info(f"The new schema is: {self.new_schema_fields}")

        column_names_new = [i["name"] for i in self.new_schema_fields]
        column_names_current = [i["name"] for i in self.current_schema_fields]

        sql_columns = []
        schema_fields_updates = []
        change_log = []

        for field in self.current_schema_fields:
            column_name = field["name"]
            column_type = field["type"]
            column_mode = field["mode"]

            if (column_name not in column_names_new) and (column_mode == "REQUIRED"):
                logging.info(
                    f"Column `{column_name}` was changed from REQUIRED to NULLABLE"
                )

                change_log.append(
                    {
                        "table_id": self.table_id,
                        "dataset_id": self.dataset_id,
                        "column_name": column_name,
                        "type_of_change": "column mode relaxation",
                    }
                )

                schema_fields_updates.append({"name": column_name, "mode": "NULLABLE"})

                sql_columns.append(f"""`{column_name}`""")

            else:
                column_type_new = self.bigQuery_mapping(
                    next(
                        (
                            i["type"]
                            for i in self.new_schema_fields
                            if i["name"] == column_name
                        ),
                        None,
                    )
                )
                if (column_type_new is not None) and (column_type_new != column_type):
                    if not self.allowed_casting(
                        column_name, column_type, column_type_new
                    ):
                        raise AirflowException(
                            f"Data type of column {column_name} cannot be changed from {column_type} to {column_type_new}"
                        )

                    logging.info(
                        f"Data type of column `{column_name}` was changed from {column_type} to {column_type_new}"
                    )

                    change_log.append(
                        {
                            "table_id": self.table_id,
                            "dataset_id": self.dataset_id,
                            "column_name": column_name,
                            "type_of_change": "data type change",
                        }
                    )

                    sql_columns.append(
                        f"""CAST(`{column_name}` AS {column_type_new}) AS `{column_name}`"""
                    )
                else:
                    sql_columns.append(f"""`{column_name}`""")

        for field in self.new_schema_fields:
            column_name = field["name"]
            column_type = field["type"]

            if column_name not in column_names_current:
                logging.info(f"{column_name} was added to the table")

                change_log.append(
                    {
                        "table_id": self.table_id,
                        "dataset_id": self.dataset_id,
                        "column_name": column_name,
                        "type_of_change": "column addition",
                    }
                )

                schema_fields_updates.append(field)

        query = f"""SELECT {",".join(sql_columns)} FROM `{self.dataset_id}.{self.table_id}`;"""

        return query, schema_fields_updates, sql_columns, change_log

    def bigQuery_mapping(self, data_type):
        mapping = {"FLOAT": "FLOAT"}

        if data_type in mapping:
            return mapping[data_type]
        else:
            return data_type

    def allowed_casting(self, column_name, current_data_type, new_data_type):
        casting = {
            "INTEGER": [
                "BOOL",
                "INTEGER",
                "NUMERIC",
                "BIGNUMERIC",
                "FLOAT64",
                "STRING",
            ],
            "NUMERIC": ["INTEGER", "NUMERIC", "BIGNUMERIC", "FLOAT64", "STRING"],
            "BIGNUMERIC": ["INTEGER", "NUMERIC", "BIGNUMERIC", "FLOAT64", "STRING"],
            "FLOAT": ["INTEGER", "NUMERIC", "BIGNUMERIC", "FLOAT64", "STRING"],
            "BOOL": ["BOOL", "INTEGER", "STRING"],
            "STRING": [
                "BOOL",
                "INTEGER",
                "NUMERIC",
                "BIGNUMERIC",
                "FLOAT64",
                "STRING",
                "BYTES",
                "DATE",
                "DATETIME",
                "TIME",
                "TIMESTAMP",
            ],
            "BYTES": ["STRING", "BYTES"],
            "DATE": ["STRING", "DATE", "DATETIME", "TIMESTAMP"],
            "DATETIME": ["STRING", "DATE", "DATETIME", "TIME", "TIMESTAMP"],
            "TIME": ["STRING", "TIME"],
            "TIMESTAMP": ["STRING", "DATE", "DATETIME", "TIME", "TIMESTAMP"],
            "ARRAY": ["ARRAY"],
            "STRUCT": ["STRUCT"],
        }

        allowed_casting = casting[current_data_type]

        if new_data_type not in allowed_casting:
            return False
        else:
            return True



from dacite import Config
from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass

from typing import List, Optional
import regex as re

from gcp_airflow_foundations.enums.schema_source_type import SchemaSourceType


@dataclass
class FullIngestionConfig:
    """
    Attributes:
        ingest_all_tables: if true, ingest all tables from source
        ingestion_name: name to use for DAG id, should correspond to the relevant partition of the source cut out by the regex expression provided
        dag_creation_mode: if "SOURCE", then one DAG per source is created. if "TABLE", one DAG per table.
        regex_table_matching: regex pattern to match tables to, if ingest_all_tables is false
    """
    ingest_all_tables: bool = False
    ingestion_name: str = ""
    dag_creation_mode: str = "TABLE"
    regex_table_pattern: str = "ANY"



from typing import Dict, Iterable, Union

from airflow.operators.branch import BaseBranchOperator
from airflow.utils import timezone
from croniter import croniter
from datetime import datetime, timedelta

import logging


class BranchOnCronOperator(BaseBranchOperator):
    """
    Branches into one of two lists of tasks depending on the current day and cron expressions
    :param follow_task_ids_if_true: task id or task ids to follow if criteria met
    :type follow_task_ids_if_true: str or list[str]
    :param follow_task_ids_if_false: task id or task ids to follow if criteria does not met
    :type follow_task_ids_if_false: str or list[str]
    :param cron_expression: cron expression to evaluate
    :type cron_expression: str
    :param use_task_execution_day: If ``True``, uses task's execution day to compare
        with is_today. Execution Date is Useful for backfilling.
        If ``False``, uses system's day of the week.
    :type use_task_execution_day: bool
    :param run_on_first_execution: If ``True``, returns true on  first time this task is run
    :type run_on_first_execution: bool
    """

    def __init__(
        self,
        *,
        follow_task_ids_if_true: Union[str, Iterable[str]],
        follow_task_ids_if_false: Union[str, Iterable[str]],
        cron_expression: str,
        use_task_execution_day: bool = True,
        run_on_first_execution: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.follow_task_ids_if_true = follow_task_ids_if_true
        self.follow_task_ids_if_false = follow_task_ids_if_false
        self.cron_expression = cron_expression
        self.use_task_execution_day = use_task_execution_day
        self.run_on_first_execution = run_on_first_execution

        if not croniter.is_valid(cron_expression):
            raise TypeError("Unsupported cron expression {}".format(cron_expression))

    def choose_branch(self, context: Dict) -> Union[str, Iterable[str]]:
        logging.info(f"context {context}")
        ti = context["ti"]
        logging.info(f"ti {ti}")

        if self.use_task_execution_day:
            now = context["execution_date"]
        else:
            now = timezone.make_naive(timezone.utcnow(), self.dag.timezone)

        first_run = ti.get_previous_ti() is None

        if first_run and self.run_on_first_execution:
            return self.follow_task_ids_if_true
        elif self.match(self.cron_expression, now):
            return self.follow_task_ids_if_true
        else:
            return self.follow_task_ids_if_false

    def match(self, cron_expression, date):
        iter = croniter(cron_expression, date)
        iter.get_prev()
        return iter.get_next(datetime) == date



from dacite import Config
from dacite.exceptions import MissingValueError
from dacite.dataclasses import DefaultValueNotFoundError

from gcp_airflow_foundations.base_class.data_source_table_config import (
    DataSourceTablesConfig,
)
from enum import Enum
import yaml
import dacite
import logging
import glob

from airflow.exceptions import AirflowException


def load_tables_config(config_path) -> DataSourceTablesConfig:
    table_config_file = open(config_path, "r").read()
    raw_config = yaml.load(table_config_file, Loader=yaml.FullLoader)

    try:
        config = dacite.from_dict(
            data_class=DataSourceTablesConfig,
            data=raw_config,
            config=Config(cast=[Enum]),
        )
        logging.info(f"Config is {config}")
    except (MissingValueError, DefaultValueNotFoundError, KeyError) as e:
        raise AirflowException(str(e))

    return config


def load_tables_config_from_dir(dir_path):
    files = glob.glob(f"{dir_path}/*.yaml")
    logging.info(f"Loading config from  {files}")
    return list(map(load_tables_config, files))



from abc import ABC, abstractmethod

from gcp_airflow_foundations.common.gcp.source_schema.gcs import read_schema_from_gcs
from gcp_airflow_foundations.common.gcp.source_schema.bq import read_schema_from_bq


class SchemaSourceConfig(ABC):
    @abstractmethod
    def schema_method(self):
        pass

    @abstractmethod
    def schema_method_arguments(self, data_source, table_config, ds):
        pass


class AutoSchemaSourceConfig(SchemaSourceConfig):
    def schema_method(self):
        return

    def schema_method_arguments(self, data_source, table_config, ds):
        return


class GCSSchemaSourceConfig(SchemaSourceConfig):
    def schema_method(self):
        return read_schema_from_gcs

    def schema_method_arguments(self, data_source, table_config, ds):
        return {
            "gcs_schema_object": data_source.schema_options.schema_object_template.format(
                table_name=table_config.table_name
            )
        }


class BQLandingZoneSchemaSourceConfig(SchemaSourceConfig):
    def schema_method(self):
        return read_schema_from_bq

    def schema_method_arguments(self, data_source, table_config, ds):
        return {
            "dataset_id": data_source.landing_zone_options.landing_zone_dataset,
            "table_id": f"{table_config.landing_zone_table_name_override}_{ds}",
            "project_id": data_source.gcp_project
        }



from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryUpdateTableSchemaOperator,
)
from operators.gcp.dlp.dlp_helpers import fields_to_policy_tags
from typing import Optional


class UpdatePolicyTagsOperator(BigQueryUpdateTableSchemaOperator):
    template_fields = ("dataset_id", "table_id", "fields", "policy_tag")

    def __init__(
        self,
        *,
        dataset_id: str,
        table_id: str,
        project_id: str,
        fields: list[str],
        policy_tag: str,
        gcp_conn_id: str = "google_cloud_default",
        delegate_to: Optional[str] = None,
        **kwargs,
    ) -> None:
        schema_fields_updates = fields_to_policy_tags(fields, policy_tag)
        super(BigQueryUpdateTableSchemaOperator, self).__init__(
            delegate_to=delegate_to,
            gcp_conn_id=gcp_conn_id,
            dataset_id=dataset_id,
            project_id=project_id,
            table_id=table_id,
            schema_fields_updates=schema_fields_updates,
            include_policy_tags=True,
            **kwargs,
        )



import json
import logging
import uuid
from datetime import datetime

from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook

from google.cloud import bigquery


class SchemaMigrationAudit:
    """
    Inserts the the audit log rows of the schema migration operations that are executed by the MigrateSchema class. The table is created if it does not exist.
    :param project_id: GCP project ID
    :type project_id: str
    :param table_id: Target table name
    :type table_id: str
    :param dataset_id: Target dataset name
    :type dataset_id: str
    :param delegate_to: The account to impersonate using domain-wide delegation of authority, if any
    :type delegate_to: str
    :param gcp_conn_id: Airflow GCP connection ID
    :type gcp_conn_id: str
    """

    def __init__(
        self,
        project_id,
        dataset_id,
        table_id="schema_migration_audit_table",
        gcp_conn_id="google_cloud_default",
        delegate_to=None,
    ):

        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.gcp_conn_id = gcp_conn_id
        self.delegate_to = delegate_to

        self.migration_id = uuid.uuid4().hex
        self.migration_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        self.bq_hook = BigQueryHook(
            gcp_conn_id=self.gcp_conn_id, delegate_to=self.delegate_to
        )

    def insert_change_log_rows(self, change_log: list):
        """
        :param change_log: The schema migration operations executed by the MigrateSchema class
        :type change_log: list
        """
        logging.info("Creating audit table in BQ.")
        self.create_audit_table()

        for i in change_log:
            i["schema_updated_at"] = self.migration_time
            i["migration_id"] = self.migration_id

        client = self.bq_hook.get_client(project_id=self.project_id)

        table_ref = client.dataset(self.dataset_id).table(self.table_id)
        table = client.get_table(table_ref)

        client.insert_rows(table, change_log)

    def create_audit_table(self):
        self.bq_hook.create_empty_table(
            project_id=self.project_id,
            dataset_id=self.dataset_id,
            table_id=self.table_id,
            schema_fields=self.__audit_table_schema_fields(),
            exists_ok=True,
        )

    def __audit_table_schema_fields(self):
        return [
            {"name": "table_id", "type": "STRING"},
            {"name": "dataset_id", "type": "STRING"},
            {"name": "schema_updated_at", "type": "TIMESTAMP"},
            {"name": "migration_id", "type": "STRING"},
            {"name": "column_name", "type": "STRING"},
            {"name": "type_of_change", "type": "STRING"},
        ]



from pydantic.dataclasses import dataclass

from typing import List, Optional


@dataclass
class FileTableConfig:
    """
    Attributes:
        directory_prefix: optional directory prefix override per table (TODO: is it optinal? what is it used for)
        metadata_file: the relative path to the metadata file within the GSC bucket
        schema_file: the (optional) relative path to the schema file for the table
        flag_file_path: the path to the optional success file
        bq_upload_option: either "BASH" or "GCS"
    """

    directory_prefix: Optional[str]
    metadata_file: Optional[str] = None
    schema_file: Optional[str] = None
    flag_file_path: Optional[str] = None
    parquet_upload_option: str = "BASH"
    allow_quoted_newlines: bool = False



from airflow.models.baseoperator import BaseOperator
from gcp_airflow_foundations.operators.api.hooks.dataform_hook import DataformHook
# from airflow.utils.decorators import apply_defaults
from typing import Any, Optional
from airflow.exceptions import AirflowException
import time
import logging


class DataformOperator(BaseOperator):
    '''
    This operator will use the custom DataformHook to POST to Dataform's ApiService_RunCreate to initiate a new dataform run.
    If no tags are provided, Dataform will run all jobs.
    After a run is created, if the validate_job_submitted parameter is set to True, it will poke Dataform for the run status for a certain amount of time. Use a separate sensor for longer jobs and set validate_job_submitted to False.

    Attributes:
    :param environment: name of the environment, e.g. production, development...
    :type environment: str
    :param schedule: the name of the schedule that will be triggered by airflow
    :type schedule: str
    :param tags: (optional) A list of tags with which the action must have in order to be run. If not set, then all actions will run.
    :type tags: list[str]
    :param submit_mode: (optional) Valid options are 'submit_job_only' (default), 'validate_job_submitted', and 'wait_for_job_to_finish'.
    :type submit_mode: str
    :param retry_period_seconds: (optional) Amount of time to wait for retry if validation job doesn't return SUCCESSFUL. If not set, then default value is 10 seconds.
    :type retry_period_seconds: int

    Instructions to prepare Dataform for the API call:
    1. create schedule in Dataform for REST API call
        a) environments.json >> under the appropriate environment (e.g. prod, dev..) click on "create new schedule"
        b) enter schedule name (e.g. composer)
        c) disable "Enable this schedule"
    2. get Dataform project ID
        easiest way is to look at the url
        E.g. https://app.dataform.co/#/<project_id>/b/<branch>/file/<file-name>
    3. generate API token:
        project settings >> API keys >> GENERATE NEW API KEY

    - Dataform documentation on using REST API: https://docs.dataform.co/dataform-web/api
    - Helpful Medium tutorial: https://medium.com/google-cloud/cloud-composer-apache-airflow-dataform-bigquery-de6e3eaabeb3

    ** Note: Schedules must be on the master branch. In Dataform, you'll have to create a branch first and then merge changes into master.
    '''

    # @apply_defaults
    def __init__(
            self,
            *,
            dataform_conn_id: str = 'dataform_default',
            project_id: str,
            environment: str,
            schedule: str,
            tags: Optional[str] = [],
            submit_mode: str = 'submit_job_only',
            retry_period_seconds: Optional[int] = 10,
            **kwargs: Any) -> None:

        super().__init__(**kwargs)
        self.dataform_conn_id = dataform_conn_id
        self.project_id = project_id
        self.environment = environment
        self.schedule = schedule
        self.tags = tags
        self.submit_mode = submit_mode
        self.retry_period_seconds = retry_period_seconds

    def execute(self, context) -> str:
        dataform_hook = DataformHook(dataform_conn_id=self.dataform_conn_id)
        run_url = dataform_hook.run_job(
            project_id=self.project_id,
            environment=self.environment,
            schedule=self.schedule,
            tags=self.tags
        )

        if self.submit_mode == 'submit_job_only':
            return run_url

        while self.submit_mode == 'validate_job_submitted':
            response = dataform_hook.check_job_status(run_url)
            logging.info(f'dataform job status: {response} run rul: {run_url}')

            if response in ('SUCCESSFUL', 'RUNNING'):
                return run_url
            else:
                time.sleep(self.retry_period_seconds)

        while self.submit_mode == 'wait_for_job_to_finish':
            response = dataform_hook.check_job_status(run_url)
            logging.info(f'dataform job status: {response} run rul: {run_url}')

            if response == 'SUCCESSFUL':
                return run_url
            elif response != 'RUNNING':
                raise AirflowException(f"Dataform ApiService Error: {response}")
            else:
                time.sleep(self.retry_period_seconds)



from airflow.providers.google.cloud.transfers.gcs_to_bigquery import (
    GCSToBigQueryOperator,
)

# from airflow.utils.decorators import apply_defaults

from gcp_airflow_foundations.source_class.schema_source_config import SchemaSourceConfig
from gcp_airflow_foundations.base_class.source_table_config import SourceTableConfig
from gcp_airflow_foundations.base_class.source_config import SourceConfig

import logging


class CustomGCSToBigQueryOperator(GCSToBigQueryOperator):
    # @apply_defaults
    def __init__(
        self,
        *,
        schema_config: SchemaSourceConfig,
        data_source: SourceConfig,
        table_config: SourceTableConfig,
        bucket: str,
        source_objects: str,
        destination_project_dataset_table: str,
        write_disposition: str,
        create_disposition: str,
        skip_leading_rows: int,
        **kwargs,
    ) -> None:
        super(CustomGCSToBigQueryOperator, self).__init__(
            bucket=bucket,
            source_objects=source_objects,
            destination_project_dataset_table=destination_project_dataset_table,
            write_disposition=write_disposition,
            create_disposition=create_disposition,
            skip_leading_rows=skip_leading_rows,
            **kwargs,
        )

        self.schema_config = schema_config
        self.data_source = data_source
        self.table_config = table_config

    def pre_execute(self, context) -> None:
        ds = context["ds"]

        schema_source_config_class = self.schema_config

        logging.info(
            f"Parsing schema for staging table using `{schema_source_config_class.__name__}`"
        )

        schema_method = schema_source_config_class().schema_method()
        schema_method_arguments = schema_source_config_class().schema_method_arguments(
            self.data_source, self.table_config, ds
        )

        if schema_method and schema_method_arguments:
            self.schema_fields = schema_method(**schema_method_arguments)
        else:
            self.schema_fields = None
            self.autodetect = True



from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional


@dataclass
class SFTPTableConfig:
    """
    Attributes:
        full_dir_download:
        date_column: optional date column name for scanning/replacing in external partitions for .PARQUET uploads
    """

    full_dir_download: bool
    date_column: Optional[str]
